@@ -74,6 +74,8 @@ struct _PanelAppletPrivate {
         int                         size_hints_len;
 
 	gboolean		    moving_focus_out;
+	gint		   	    focusable_child;
+	guint			    hierarchy_changed_id;
 };
 
 static GObjectClass *parent_class;
@@ -406,57 +408,143 @@ panel_applet_finalize (GObject *object)
 	parent_class->finalize (object);
 }
 
-static void
-panel_applet_menu_position (GtkMenu  *menu,
-			    gint     *x,
-			    gint     *y,
-			    gboolean *push_in,
-			    gpointer  data)
+static GtkWidget*
+panel_applet_container_has_focusable_child (GtkWidget *widget)
 {
-	GtkWidget *w = data;
-	GtkRequisition requisition;
-	gint wx, wy;
+	GtkContainer *container;
+	GtkWidget *child;
+	GList *list;
+	GList *t;
+	GtkWidget *retval = NULL;
 
-	g_return_if_fail (w != NULL);
+	container = GTK_CONTAINER (widget);
+	list = gtk_container_get_children (container);
+
+	for (t = list; t; t = t->next) {
+		child = GTK_WIDGET (t->data);
+		if (GTK_WIDGET_CAN_FOCUS (child)) {
+			retval = child;
+			break;
+		} else if (GTK_IS_CONTAINER (child)) {
+			retval = panel_applet_container_has_focusable_child (child);
+			if (retval)
+				break;
+		}
+	}
+	g_list_free (list);
+	return retval;	
+}
+
+static void
+panel_applet_hierarchy_changed_cb (GtkWidget *widget)
+{
+	PanelApplet *applet = PANEL_APPLET (widget);
+
+	applet->priv->focusable_child = -1;
+}
+
+static gboolean
+panel_applet_has_focusable_child (PanelApplet *applet)
+{
+	if (applet->priv->focusable_child == -1) {
+		GtkWidget *focusable_child;
+
+		if (!applet->priv->hierarchy_changed_id)
+			applet->priv->hierarchy_changed_id = 
+				g_signal_connect (applet, "hierarchy-changed", 
+				  		  G_CALLBACK (panel_applet_hierarchy_changed_cb),
+				  		  NULL);
+
+		focusable_child = panel_applet_container_has_focusable_child (GTK_WIDGET (applet));
+		applet->priv->focusable_child = (focusable_child != NULL);
+	}
+	return  (applet->priv->focusable_child != 0);
+}
+
+static void
+panel_applet_position_menu (GtkMenu   *menu,
+			    int       *x,
+			    int       *y,
+			    gboolean  *push_in,
+			    GtkWidget *widget)
+{
+	PanelApplet    *applet;
+	GtkRequisition  requisition;
+	GdkScreen      *screen;
+	int             menu_x = 0;
+	int             menu_y = 0;
+
+	g_return_if_fail (PANEL_IS_APPLET (widget));
+
+	applet = PANEL_APPLET (widget);
+
+	screen = gtk_widget_get_screen (widget);
 
 	gtk_widget_size_request (GTK_WIDGET (menu), &requisition);
 
-	gdk_window_get_origin (w->window, &wx, &wy);
-	/*
-         * Make sure that the popup position is in the panel
-	 * as the menu may be popped up by a keystroke
-	 */
-	if (*x < wx)
-		*x = wx;
-	else if (*x > wx + w->allocation.width)
-		*x = wx + w->allocation.width;
+	gdk_window_get_origin (widget->window, &menu_x, &menu_y);
 
-	if (*x + requisition.width > gdk_screen_width())
-		*x = gdk_screen_width() - requisition.width;
+	menu_x += widget->allocation.x;
+	menu_y += widget->allocation.y;
 
-	if (*y < wy)
-		*y = wy;
-	 else if (*y > wy + w->allocation.height)
-		*y = wy + w->allocation.height;
+	if (applet->priv->orient == PANEL_APPLET_ORIENT_UP ||
+	    applet->priv->orient == PANEL_APPLET_ORIENT_DOWN) {
+		if (menu_y > gdk_screen_get_height (screen) / 2)
+			menu_y -= requisition.height;
+		else
+			menu_y += widget->allocation.height;
+	} else  {
+		if (menu_x > gdk_screen_get_width (screen) / 2)
+			menu_x -= requisition.width;
+		else
+			menu_x += widget->allocation.width;
 
-	if (*y + requisition.height > gdk_screen_height())
-		*y = gdk_screen_height() - requisition.height;
+	}
 
+	*x = menu_x;
+	*y = menu_y;
 	*push_in = TRUE;
 }
 
+static gboolean
+panel_applet_can_focus (GtkWidget *widget)
+{
+	/*
+	 * A PanelApplet widget can focus if it has a tooltip or it does 
+	 * not have any focusable children.
+	 */
+	if (gtk_tooltips_data_get (widget))
+		return TRUE;
+
+	if (!PANEL_IS_APPLET (widget))
+		return FALSE;
+
+	return !panel_applet_has_focusable_child (PANEL_APPLET (widget));
+}
+
 static gboolean
 panel_applet_button_press (GtkWidget      *widget,
 			   GdkEventButton *event)
 {
 	PanelApplet *applet = PANEL_APPLET (widget);
 
+	if (!panel_applet_has_focusable_child (applet)) {
+		if (!GTK_WIDGET_HAS_FOCUS (widget)) {
+			GTK_WIDGET_SET_FLAGS (widget, GTK_CAN_FOCUS);
+			gtk_widget_grab_focus (widget);
+		}
+	}
+
 	if (event->button == 1)
 		return TRUE;
 	else if (event->button == 3) {
-		bonobo_control_do_popup (applet->priv->control, 
-					 event->button,
-					 event->time);
+		bonobo_control_do_popup_full (
+				applet->priv->control, 
+				NULL, NULL,
+				(GtkMenuPositionFunc) panel_applet_position_menu,
+				applet,
+				event->button,
+				event->time);
 		return TRUE;
 	}
 
@@ -469,13 +557,12 @@ _panel_applet_popup_menu (PanelApplet *applet,
 			  guint32 time)			  
 {
 	bonobo_control_do_popup_full (applet->priv->control, NULL, NULL,
-				      panel_applet_menu_position,
-				      GTK_WIDGET (applet), button,
-				      time);
+				      (GtkMenuPositionFunc) panel_applet_position_menu,
+				      applet, button, time);
 	return TRUE;
 }
 
-gboolean
+static gboolean
 panel_applet_popup_menu (PanelApplet *applet)
 {
 	return _panel_applet_popup_menu (applet, 3, GDK_CURRENT_TIME);
@@ -484,52 +571,56 @@ panel_applet_popup_menu (PanelApplet *applet)
 static void
 panel_applet_size_request (GtkWidget *widget, GtkRequisition *requisition)
 {
-	PanelApplet *applet = PANEL_APPLET (widget);
-	GtkBin *bin = GTK_BIN (widget);
-	gint focus_width, focus_pad;
+	int focus_width = 0;
 
-	requisition->width = GTK_CONTAINER (widget)->border_width * 2;
-	requisition->height = GTK_CONTAINER (widget)->border_width * 2;
+	GTK_WIDGET_CLASS (parent_class)->size_request (widget, requisition);
 
+	if (!panel_applet_can_focus (widget))
+		return;
+
+	/*
+	 * We are deliberately ignoring focus-padding here to
+	 * save valuable panel real estate.
+	 */
 	gtk_widget_style_get (widget,
 			      "focus-line-width", &focus_width,
-			      "focus-padding", &focus_pad,
 			      NULL);
 
-	if (bin->child && GTK_WIDGET_VISIBLE (bin->child)) {
-		GtkRequisition child_requisition;
-
-		gtk_widget_size_request (bin->child, &child_requisition);
-
-		requisition->width += child_requisition.width;
-		requisition->height += child_requisition.height;
-	}
-	requisition->width += 2 * (focus_pad + 2 * focus_width);
-	requisition->height += 2 * (focus_pad + 2 * focus_width);
+	requisition->width  += 2 * focus_width;
+	requisition->height += 2 * focus_width;
 }
 
 static void
-panel_applet_size_allocate (GtkWidget *widget, GtkAllocation *allocation)
+panel_applet_size_allocate (GtkWidget     *widget,
+			    GtkAllocation *allocation)
 {
-	PanelApplet *applet = PANEL_APPLET (widget);
-	gint focus_width, focus_pad;
-	gint x, y, width, height;
-	gint border_width;
-	GtkBin *bin;
-	GtkAllocation child_allocation;
+	GtkAllocation  child_allocation;
+	GtkBin        *bin;
+	int            border_width;
+	int            focus_width = 0;
 
+	if (!panel_applet_can_focus (widget)) {
+		GTK_WIDGET_CLASS (parent_class)->size_allocate (widget, allocation);
+		return;
+	}
+
+	/*
+	 * We are deliberately ignoring focus-padding here to
+	 * save valuable panel real estate.
+	 */
 	gtk_widget_style_get (widget,
 			      "focus-line-width", &focus_width,
-			      "focus-padding", &focus_pad,
 			      NULL);
+
 	border_width = GTK_CONTAINER (widget)->border_width;
 
 	widget->allocation = *allocation;
 	bin = GTK_BIN (widget);
 
- 	child_allocation.x = focus_width + 2 * focus_pad;
- 	child_allocation.y = focus_width + 2 * focus_pad;
-	child_allocation.width = MAX (allocation->width - border_width * 2, 0);
+ 	child_allocation.x = focus_width;
+ 	child_allocation.y = focus_width;
+
+	child_allocation.width  = MAX (allocation->width  - border_width * 2, 0);
 	child_allocation.height = MAX (allocation->height - border_width * 2, 0);
 
 	if (GTK_WIDGET_REALIZED (widget))
@@ -538,43 +629,51 @@ panel_applet_size_allocate (GtkWidget *widget, GtkAllocation *allocation)
 					allocation->y + GTK_CONTAINER (widget)->border_width,
 					child_allocation.width,
 					child_allocation.height);
-	child_allocation.width = MAX (child_allocation.width - 2 * (focus_width + 2 * focus_pad), 0);
-	child_allocation.height = MAX (child_allocation.height - 2 * (focus_width + 2 * focus_pad), 0);
+
+	child_allocation.width  = MAX (child_allocation.width  - 2 * focus_width, 0);
+	child_allocation.height = MAX (child_allocation.height - 2 * focus_width, 0);
 
 	if (bin->child)
 		gtk_widget_size_allocate (bin->child, &child_allocation);
 }
 
-
 static gboolean
 panel_applet_expose (GtkWidget      *widget,
 		     GdkEventExpose *event) 
 {
+	int border_width;
+	int focus_width = 0;
+	int x, y, width, height;
+
 	g_return_val_if_fail (PANEL_IS_APPLET (widget), FALSE);
 	g_return_val_if_fail (event != NULL, FALSE);
 
 	GTK_WIDGET_CLASS (parent_class)->expose_event (widget, event);
 
-        if (GTK_WIDGET_HAS_FOCUS (widget)) {
-		gint focus_width, focus_pad;
-		gint x, y, width, height;
-		gint border_width;
+        if (!GTK_WIDGET_HAS_FOCUS (widget))
+		return FALSE;
 
-		gtk_widget_style_get (widget,
-				      "focus-line-width", &focus_width,
-				      "focus-padding", &focus_pad,
-				      NULL);
-		border_width = GTK_CONTAINER (widget)->border_width;
-		x = widget->allocation.x + focus_pad;
-		y = widget->allocation.y + focus_pad;
-		width = widget->allocation.width - 2 * border_width - (focus_width + 2 * focus_pad);
-		height = widget->allocation.height - 2 * border_width - (focus_width +  2 * focus_pad);
+	/*
+	 * We are deliberately ignoring focus-padding here to
+	 * save valuable panel real estate.
+	 */
+	gtk_widget_style_get (widget,
+			      "focus-line-width", &focus_width,
+			      NULL);
+
+	border_width = GTK_CONTAINER (widget)->border_width;
+
+	x = widget->allocation.x;
+	y = widget->allocation.y;
+
+	width  = widget->allocation.width  - 2 * border_width;
+	height = widget->allocation.height - 2 * border_width;
+
+	gtk_paint_focus (widget->style, widget->window,
+			 GTK_WIDGET_STATE (widget),
+			 &event->area, widget, "panel_applet",
+			 x, y, width, height);
 
-		gtk_paint_focus (widget->style, widget->window,
-                                 GTK_WIDGET_STATE (widget),
-                                 &event->area, widget, "panel_applet",
-                                 x, y, width, height);
-	}
 	return FALSE;
 }                
 
@@ -589,8 +688,14 @@ panel_applet_focus (GtkWidget        *widget,
 	g_return_val_if_fail (PANEL_IS_APPLET (widget), FALSE);
 
 	applet = PANEL_APPLET (widget);
-	if (applet->priv->moving_focus_out)
+	if (applet->priv->moving_focus_out) {
+		/*
+		 * Applet will retain focus if there is nothing else on the
+		 * panel to get focus
+		 */
+		applet->priv->moving_focus_out = FALSE;
 		return FALSE;
+	}
 
 	previous_focus_child = GTK_CONTAINER (widget)->focus_child;
 	 if (!previous_focus_child && !GTK_WIDGET_HAS_FOCUS (widget)) {
@@ -732,15 +837,18 @@ panel_applet_handle_background_string (PanelApplet  *applet,
 				       GdkColor     *color,
 				       GdkPixmap   **pixmap)
 {
-	gchar **elements;
+	PanelAppletBackgroundType   retval;
+	char                      **elements;
+
+	retval = PANEL_NO_BACKGROUND;
 
 	if (!applet->priv->background)
-		return PANEL_NO_BACKGROUND;
+		return retval;
 
 	elements = g_strsplit (applet->priv->background, ":", -1);
 
 	if (elements [0] && !strcmp (elements [0], "none" )) {
-		return PANEL_NO_BACKGROUND;
+		retval = PANEL_NO_BACKGROUND;
 		
 	} else if (elements [0] && !strcmp (elements [0], "color")) {
 		g_return_val_if_fail (color != NULL, PANEL_NO_BACKGROUND);
@@ -752,7 +860,7 @@ panel_applet_handle_background_string (PanelApplet  *applet,
 			return PANEL_NO_BACKGROUND;
 		}
 
-		return PANEL_COLOR_BACKGROUND;
+		retval = PANEL_COLOR_BACKGROUND;
 
 	} else if (elements [0] && !strcmp (elements [0], "pixmap")) {
 		GdkNativeWindow pixmap_id;
@@ -775,11 +883,13 @@ panel_applet_handle_background_string (PanelApplet  *applet,
 			return PANEL_NO_BACKGROUND;
 		}
 
-		return PANEL_PIXMAP_BACKGROUND;
+		retval = PANEL_PIXMAP_BACKGROUND;
 	} else
 		g_warning (_("Unknown background type received"));
 
-	return PANEL_NO_BACKGROUND;
+	g_strfreev (elements);
+
+	return retval;
 }
 
 
@@ -789,7 +899,13 @@ panel_applet_handle_background_string (PanelApplet  *applet,
  * @color: A #GdkColor to be filled in.
  * @pixmap: Returned #GdkPixmap
  *
- * Returns the current background type and fills in the relevant
+ * Returns the current background type. If the background
+ * type is %PANEL_NO_BACKGROUND both @color and @pixmap will
+ * be unaffected. If the background type is %PANEL_COLOR_BACKGROUND
+ * then @color will contain the current panel background colour.
+ * If the background type is %PANEL_PIXMAP_BACKGROUND, @pixmap will
+ * contain a pointer to a #GdkPixmap which is a copy of the applet's
+ * portion of the panel's background pixmap.
  * 
  * Return value: a #PanelAppletOrient value.
  */
@@ -1286,14 +1402,15 @@ panel_applet_setup (PanelApplet *applet)
 
 	g_signal_connect (applet, "popup_menu",
 			  G_CALLBACK (panel_applet_popup_menu), NULL);
+
+	priv->focusable_child = -1;
+	priv->hierarchy_changed_id  = 0;
 }
 
 /**
  * panel_applet_new:
- * @widget: The widget the contains all the widgetry the applet
- *          wishes to expose.
  *
- * Creates a new #PanelApplet which exposes @widget.
+ * Creates a new #PanelApplet.
  *
  * Return value: A #GtkWidget on success, %NULL on failure.
  */
@@ -1335,110 +1452,6 @@ panel_applet_callback_data_free (PanelAppletCallBackData *data)
 	g_free (data);
 }
 
-/*
- *   Time we're prepared to wait without a ControlFrame
- * before terminating the Control. This can happen if the
- * panel activates us but crashes before the set_frame.
- *
- * NB. if we don't get a frame in 30 seconds, something
- * is badly wrong, or Gnome performance needs improving
- * markedly !
- */
-#define PANEL_APPLET_NEVER_GOT_FRAME_TIMEOUT 30 * 1000 /* ms */
-
-/* Track the living controls */
-static GSList *active_controls = NULL;
-
-static gboolean
-panel_applet_idle_quit (gpointer data)
-{
-	if (!active_controls)
-		bonobo_main_quit ();
-
-	return FALSE;
-}
-
-static void
-panel_applet_cnx_broken_callback (BonoboControl *control)
-{
-	active_controls = g_slist_remove (active_controls, control);
-	if (!active_controls)
-		g_idle_add (panel_applet_idle_quit, NULL);
-}
-
-static gboolean
-panel_applet_never_got_frame_timeout (gpointer user_data)
-{
-	g_warning ("Never got frame, panel died - abnormal exit condition");
-
-	panel_applet_cnx_broken_callback (user_data);
-	
-	return FALSE;
-}
-
-static void
-panel_applet_set_frame_callback (BonoboControl *control,
-				 gpointer       user_data)
-{
-	Bonobo_ControlFrame remote_frame;
-
-	remote_frame = bonobo_control_get_control_frame (control, NULL);
-
-	if (remote_frame != CORBA_OBJECT_NIL) {
-		ORBitConnectionStatus status;
-
-		g_source_remove (GPOINTER_TO_UINT (user_data));
-
-		status = ORBit_small_get_connection_status (remote_frame);
-
-		/* Only track out of proc controls */
-		if (status != ORBIT_CONNECTION_IN_PROC) {
-			active_controls = g_slist_prepend (active_controls,
-							   control);
-
-			g_signal_connect_closure (
-				ORBit_small_get_connection (remote_frame),
-				"broken",
-				g_cclosure_new_object_swap (
-					G_CALLBACK (panel_applet_cnx_broken_callback),
-					G_OBJECT (control)),
-				FALSE);
-			g_signal_connect (
-				control, "destroy",
-				G_CALLBACK (panel_applet_cnx_broken_callback),
-				NULL);
-		}
-	}
-}
-
-/**
- * panel_applet_instrument_for_failure:
- * @control: the control.
- * 
- *   Don't read this method - just believe in it; it'll
- * move inside libbonoboui in due course.
- **/
-static void
-panel_applet_instrument_for_failure (BonoboControl *control)
-{
-	guint no_frame_timeout_id;
-
-	no_frame_timeout_id = g_timeout_add (
-		PANEL_APPLET_NEVER_GOT_FRAME_TIMEOUT,
-		panel_applet_never_got_frame_timeout,
-		control);
-	g_signal_connect_closure (
-		control, "destroy",
-		g_cclosure_new_swap (
-			G_CALLBACK (g_source_remove_by_user_data),
-			control, NULL),
-		0);
-	g_signal_connect (
-		control, "set_frame",
-		G_CALLBACK (panel_applet_set_frame_callback),
-		GUINT_TO_POINTER (no_frame_timeout_id));
-}
-
 static BonoboObject *
 panel_applet_factory_callback (BonoboGenericFactory    *factory,
 			       const char              *iid,
@@ -1453,11 +1466,18 @@ panel_applet_factory_callback (BonoboGenericFactory    *factory,
 	applet->priv->iid     = g_strdup (iid);
 	applet->priv->closure = g_closure_ref (data->closure);
 
-	panel_applet_instrument_for_failure (applet->priv->control);
-	
+	bonobo_control_life_instrument (applet->priv->control);
+
 	return BONOBO_OBJECT (applet->priv->control);
 }
 
+static void
+panel_applet_all_controls_dead (void)
+{
+	if (!bonobo_control_life_get_count())
+		bonobo_main_quit ();
+}
+
 /**
  * panel_applet_factory_main_closure:
  * @iid: The bonobo-activation iid of the factory.
@@ -1485,6 +1505,8 @@ panel_applet_factory_main_closure (const gchar *iid,
 	bindtextdomain (GETTEXT_PACKAGE, GNOMELOCALEDIR);
 	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
 
+	bonobo_control_life_set_callback (panel_applet_all_controls_dead);
+
 	closure = bonobo_closure_store (closure, panel_applet_marshal_BOOLEAN__STRING);
 
 	data = panel_applet_callback_data_new (applet_type, closure);
